/* A simple server in the internet domain using TCP
   The port number is passed as an argument 
   This version runs forever, forking off a separate 
   process for each connection
   gcc server2.c -lsocket
*/
#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include<stdlib.h>
#include<unistd.h>
#include<strings.h>
#include<string.h>
#include<pthread.h>	
#include "header_files/toxml.h"
#include "header_files/KVCache.h"
#include "header_files/threadpool.h"


/*JOB QUEUE POINTER*/
jobs_queue *job_head=NULL;
jobs_queue *job_tail=NULL;
tpool_t *tpool;

/*CACHE POINTERS*/
kvcache_data **t;
int *repl;

/*LOCK FOR INSERTING IN JOB QUEUE*/
pthread_mutex_t job_q_lock;
/*LOCK FOR CACHE PER SETS*/
pthread_mutex_t* cache_lock;

/* FUNCION PROTOTYPE FOR SERVING THE CLIENT */
void *serve(void*); 




/* FUNCTION FOR HANDLING ERRORS*/
void error(char *msg)
{
    perror(msg);
    exit(1);
}


int main(int argc, char *argv[])
{
     socklen_t sockfd, newsockfd, clilen;
     struct sockaddr_in serv_addr, cli_addr;

     /*Cache related variables initialization*/
     t= init_cache();
     repl=rep_line_per_set();
     cache_lock=lock_initialize();

     /*Creating threadpool*/
     tpool=tpool_create(10,t,repl,cache_lock,job_q_lock);
     /*Initializing the lock for inserting and reading in job_queue*/
	   pthread_mutex_init(&job_q_lock,NULL);

     sockfd = socket(AF_INET, SOCK_STREAM, 0);
     if (sockfd < 0) 
        error("ERROR opening socket");
     bzero((char *) &serv_addr, sizeof(serv_addr));
     serv_addr.sin_family = AF_INET;
     serv_addr.sin_addr.s_addr = INADDR_ANY;
     serv_addr.sin_port = htons(8080);
     if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0) 
              error("ERROR on binding");
     listen(sockfd,5);
     clilen = sizeof(cli_addr);
     printf("Server started\n");
     while (1) {
         newsockfd = accept(sockfd,(struct sockaddr *) &cli_addr, &clilen);
      
         pthread_t thr;
         int *arg;
         arg=malloc(sizeof(*arg));
         *arg=newsockfd;
         pthread_create(&thr,NULL,serve,arg);
     } /* end of while */
     return 0; /* we never get here */
}


/******** serve() *********************
 There is a separate instance of this function 
 for each connection.  It handles all communication
 once a connnection has been established.
 *****************************************/
void *  serve (void* socket)
{

   int sock;
   int n;
   long tid=pthread_self();
   sock=*((int*) socket);
   char buffer[300000];
     
   while(1){
   //zeroing all the value in buffer
   bzero(buffer,300000);
   
   //reading the socket
   n = read(sock,buffer,300000);
   if (n < 0) error("ERROR reading from socket");
   else if (n==0) {
    printf("Client has been closed,Killing the thread\n");
    pthread_cancel(tid);
   }
   else{
   //Creating a structure to add in linked list of job queue
   jobs_queue *job=(jobs_queue *)malloc((sizeof(jobs_queue)));
   job->fd=sock;
   job->next=NULL;

    // printf("%d\n",job->fd );
   fromXML(job->key,job->value,job->messagetype,job->message,buffer);
   if(strlen(job->key)>256){
    char resp[300000];
    toXML(resp,"","","resp","Oversized key");
    write(job->fd,resp,strlen(resp));
    continue;
   }
   else if(strlen(job->key)==0){
    char resp[300000];
    toXML(resp,"","","resp","Undersized key");
    write(job->fd,resp,strlen(resp));
    continue;
   }
   if(strlen(job->value)>(256*1024)){
    char resp[300000];
    toXML(resp,"","","resp","Oversized Value");
    write(job->fd,resp,strlen(resp));
  continue;
   }
  
   if(!(!strcmp(job->messagetype,"getreq") || !strcmp(job->messagetype,"putreq") || !strcmp(job->messagetype,"delreq"))){
    char resp[300000];
    toXML(resp,"","","resp","Wrong Command");
    write(job->fd,resp,strlen(resp));
  continue;
   }
   //Adding to the chain in Job Queue
   pthread_mutex_lock(&job_q_lock);
   printf("New Request Arrived by Client No. %d\n",job->fd);
    if (tpool->head == NULL) {
        tpool->head = job;
        tpool->tail  = job;
    } else {
        tpool->tail->next = job;
        tpool->tail       = job;
    }

    pthread_cond_broadcast(&(tpool->work_cond));
   //printf("job_added:unlock\n");
   pthread_cond_signal(&tpool->work_cond);
   pthread_mutex_unlock(&job_q_lock);
   //process_work(job);
   }
   }
}
